# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import pickle

from .utils import split_path


class Cache(dict):
    def __init__(self, django_folder, cache_path):
        self.django_folder = django_folder
        project_name = split_path(django_folder)[-2]
        self.filename = os.path.join(cache_path, "{0}.pickle".format(project_name))

        if not os.path.exists(os.path.dirname(self.filename)):
            os.makedirs(os.path.dirname(self.filename))

        super(Cache, self).__init__()

    def load(self):
        try:
            with open(self.filename, "rb") as f:
                tmp_dict = pickle.load(f)
                self.update(tmp_dict)
        except IOError:
            pass

    def save(self):
        with open(self.filename, "wb") as f:
            pickle.dump(self, f, protocol=2)

    def md5(self, path):
        hash_md5 = hashlib.md5()
        with open(path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil

_BASE_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'tests/')
_FIXTURES_FOLDER = os.path.join(_BASE_DIR, 'test_project_fixtures/')

ADD_NOT_NULL_COLUMN_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_add_not_null_column/')
CREATE_TABLE_WITH_NOT_NULL_COLUMN_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_create_table_with_not_null_column/')
DROP_COLUMN_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_drop_column/')
RENAME_COLUMN_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_rename_column/')
ALTER_COLUMN_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_alter_column')
RENAME_TABLE_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_rename_table/')
ADD_NOT_NULL_COLUMN_FOLLOWED_BY_DEFAULT_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_add_not_null_column_followed_by_default/')
MULTI_COMMIT_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_multi_commit/')
CORRECT_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_correct_project/')
DELETED_MIGRATION_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_deleted_migration_project/')
IGNORE_MIGRATION_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_project_ignore_migration/')

NOT_DJANGO_GIT_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_git_project/')
NOT_GIT_DJANGO_PROJECT = os.path.join(
    _FIXTURES_FOLDER, 'test_django_without_git_project/')

ALL_GIT_PROJECTS = (
    NOT_DJANGO_GIT_PROJECT,
    MULTI_COMMIT_PROJECT,
    DELETED_MIGRATION_PROJECT
)


def prepare_git_project(path):
    """Copy the git/ folder to .git/ so it is
    actually versioned"""
    src_git = os.path.join(path, 'git/')
    dest_git = os.path.join(path, '.git/')
    try:
        shutil.copytree(src_git, dest_git)
    except OSError:
        pass  # .git/ might already exist


def clear_all_git_projects():
    """Delete all .git/ folders in test projects"""
    for project_path in ALL_GIT_PROJECTS:
        git_path = os.path.join(project_path, '.git/')
        shutil.rmtree(git_path, ignore_errors=True)



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

from django_migration_linter import MigrationLinter
from django_migration_linter.migration import Migration
from tests import fixtures


class TestLinterFunctions(unittest.TestCase):
    def test_get_sql(self):
        project_path = fixtures.ADD_NOT_NULL_COLUMN_PROJECT
        linter = MigrationLinter(project_path)
        sql_statements = linter.get_sql('test_app', '0001')
        self.assertEqual(len(sql_statements), 6)
        self.assertEqual(sql_statements[0], 'BEGIN;')
        self.assertEqual(sql_statements[-1], 'COMMIT;')

    def test_has_errors(self):
        project_path = fixtures.MULTI_COMMIT_PROJECT
        linter = MigrationLinter(project_path)
        self.assertFalse(linter.has_errors)

        m = Migration(os.path.join(project_path, 'test_app/migrations/0001_create_table.py'))
        linter.lint_migration(m)
        self.assertFalse(linter.has_errors)

        m = Migration(os.path.join(project_path, 'test_app/migrations/0002_add_new_not_null_field.py'))
        linter.lint_migration(m)
        self.assertTrue(linter.has_errors)

        m = Migration(os.path.join(project_path, 'test_app/migrations/0001_create_table.py'))
        linter.lint_migration(m)
        self.assertTrue(linter.has_errors)

    def test_linter_creation(self):
        with self.assertRaises(ValueError):
            MigrationLinter('foo/')
        with self.assertRaises(ValueError):
            MigrationLinter('/dev/null')
        with self.assertRaises(ValueError):
            MigrationLinter(fixtures.NOT_DJANGO_GIT_PROJECT)

    def test_ignore_migration_include_apps(self):
        linter = MigrationLinter(
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT,
            include_apps=('test_app2',))
        self.assertTrue(linter.should_ignore_migration('test_app1', '0001'))
        self.assertFalse(linter.should_ignore_migration('test_app2', '0001'))

    def test_ignore_migration_exclude_apps(self):
        linter = MigrationLinter(
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT,
            exclude_apps=('test_app1',))
        self.assertTrue(linter.should_ignore_migration('test_app1', '0001'))
        self.assertFalse(linter.should_ignore_migration('test_app2', '0001'))

    def test_ignore_migration_name_contains(self):
        linter = MigrationLinter(
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT,
            ignore_name_contains='foo')
        self.assertTrue(linter.should_ignore_migration('test_app', '0001_foo'))
        self.assertFalse(linter.should_ignore_migration('test_app', '0002_bar'))

    def test_ignore_migration_full_name(self):
        linter = MigrationLinter(
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT,
            ignore_name=('0001_foo',))
        self.assertTrue(linter.should_ignore_migration('test_app', '0001_foo'))
        self.assertFalse(linter.should_ignore_migration('test_app', '0002_bar'))

    def test_gather_all_migrations(self):
        linter = MigrationLinter(fixtures.CORRECT_PROJECT)
        migrations = linter._gather_all_migrations()
        self.assertEqual(len(migrations), 4)
        self.assertEqual(
            sorted([(m.app_name, m.name) for m in migrations]),
            sorted([
                ("test_app1", "0001_initial"),
                ("test_app1", "0002_a_new_null_field"),
                ("test_app2", "0001_foo"),
                ("test_app3", "0001_initial"),
            ])
        )



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import logging
import os
import re
from subprocess import Popen, PIPE
import sys

from .cache import Cache
from .constants import DEFAULT_CACHE_PATH, MIGRATION_FOLDER_NAME
from .migration import Migration
from .utils import is_directory, is_django_project, clean_bytes_to_str
from .sql_analyser import analyse_sql_statements

logger = logging.getLogger(__name__)


class MigrationLinter(object):
    def __init__(self, project_path, **kwargs):
        # Verify correctness
        if not is_directory(project_path):
            raise ValueError(
                "The given path {0} does not seem to be a directory.".format(
                    project_path
                )
            )
        if not is_django_project(project_path):
            raise ValueError(
                ("The given path {0} does not " "seem to be a django project.").format(
                    project_path
                )
            )

        # Store parameters and options
        self.django_path = project_path
        self.ignore_name_contains = kwargs.get("ignore_name_contains", None)
        self.ignore_name = kwargs.get("ignore_name", None) or tuple()
        self.include_apps = kwargs.get("include_apps", None)
        self.exclude_apps = kwargs.get("exclude_apps", None)
        self.database = kwargs.get("database", None) or "default"
        self.cache_path = kwargs.get("cache_path", None) or DEFAULT_CACHE_PATH
        self.no_cache = kwargs.get("no_cache", None) or False

        self.python_exe = (
            "{0}/bin/{1}".format(sys.prefix, "python")
            if hasattr(sys, "real_prefix")
            else "python"
        )

        # Initialise counters
        self.nb_valid = 0
        self.nb_ignored = 0
        self.nb_erroneous = 0
        self.nb_total = 0

        # Initialise cache. Read from old, write to new to prune old entries.
        self.old_cache = Cache(self.django_path, self.cache_path)
        self.new_cache = Cache(self.django_path, self.cache_path)
        if not self.no_cache:
            self.old_cache.load()

    def lint_migration(self, migration):
        app_name = migration.app_name
        migration_name = migration.name
        print("({0}, {1})... ".format(app_name, migration_name), end="")
        self.nb_total += 1

        md5hash = self.old_cache.md5(migration.abs_path)
        if md5hash in self.old_cache:
            if self.old_cache[md5hash]["result"] == "IGNORE":
                print("IGNORE (cached)")
                self.nb_ignored += 1
            elif self.old_cache[md5hash]["result"] == "OK":
                print("OK (cached)")
                self.nb_valid += 1
            else:
                print("ERR (cached)")
                self.nb_erroneous += 1

            if "errors" in self.old_cache[md5hash]:
                self.print_errors(self.old_cache[md5hash]["errors"])

            self.new_cache[md5hash] = self.old_cache[md5hash]
            return

        if self.should_ignore_migration(app_name, migration_name):
            print("IGNORE")
            self.nb_ignored += 1
            return

        sql_statements = self.get_sql(app_name, migration_name)
        analysis_result = analyse_sql_statements(sql_statements)
        errors = analysis_result["errors"]

        if analysis_result["ignored"]:
            print("IGNORE")
            self.new_cache[md5hash] = {"result": "IGNORE"}
            self.nb_ignored += 1
            return

        if not errors:
            print("OK")
            self.new_cache[md5hash] = {"result": "OK"}
            self.nb_valid += 1
            return

        print("ERR")
        self.new_cache[md5hash] = {"result": "ERR", "errors": errors}
        self.nb_erroneous += 1
        self.print_errors(errors)

    @staticmethod
    def print_errors(errors):
        for err in errors:
            error_str = "\t{0}".format(err["err_msg"])
            if err["table"]:
                error_str += " (table: {0}".format(err["table"])
                if err["column"]:
                    error_str += ", column: {0}".format(err["column"])
                error_str += ")"
            print(error_str)

    def lint_all_migrations(self, git_commit_id=None):
        # Collect migrations
        if git_commit_id:
            migrations = self._gather_migrations_git(git_commit_id)
        else:
            migrations = self._gather_all_migrations()

        # Lint those migrations
        for m in migrations:
            self.lint_migration(m)

        if not self.no_cache:
            self.new_cache.save()

    def print_summary(self):
        print("*** Summary:")
        print(
            (
                "Valid migrations: {1}/{0} - "
                "erroneous migrations: {2}/{0} - "
                "ignored migrations: {3}/{0}"
            ).format(self.nb_total, self.nb_valid, self.nb_erroneous, self.nb_ignored)
        )

    @property
    def has_errors(self):
        return self.nb_erroneous > 0

    def get_sql(self, app_name, migration_name):
        """ It would be much faster to call the
        command directly from the code using
        call_command(), but requires the code
        to setup django (by calling django.setup())
        and set the DJANGO_SETTINGS_MODULE.
        However, django is global and doesn't allow
        multiple linter instances to exist at the same time.
        (because they would all just lint the same django project)
        Even if calling a shell is slow and ugly, for now,
        it allows to seperate the instances correctly.
        """
        sqlmigrate_command = (
            "cd {0} && {1} manage.py sqlmigrate {2} {3} " "--database {4}"
        ).format(
            self.django_path, self.python_exe, app_name, migration_name, self.database
        )
        logger.info("Executing {0}".format(sqlmigrate_command))
        sqlmigrate_process = Popen(
            sqlmigrate_command, shell=True, stdout=PIPE, stderr=PIPE
        )

        sql_statements = []
        for line in map(clean_bytes_to_str, sqlmigrate_process.stdout.readlines()):
            sql_statements.append(line)
        sqlmigrate_process.wait()
        if sqlmigrate_process.returncode != 0:
            _, err = sqlmigrate_process.communicate()
            raise RuntimeError(
                "sqlmigrate command failed {0}".format(err.decode("utf-8"))
            )
        logger.info("Found {0} sql migration lines".format(len(sql_statements)))
        return sql_statements

    def _gather_migrations_git(self, git_commit_id):
        migrations = []
        # Get changes since specified commit
        git_diff_command = (
            "cd {0} && git diff --name-only --diff-filter=A {1}"
        ).format(self.django_path, git_commit_id)
        logger.info("Executing {0}".format(git_diff_command))
        diff_process = Popen(git_diff_command, shell=True, stdout=PIPE, stderr=PIPE)
        for line in map(clean_bytes_to_str, diff_process.stdout.readlines()):
            # Only gather lines that include added migrations
            if (
                re.search(r"\/{0}\/.*\.py".format(MIGRATION_FOLDER_NAME), line)
                and "__init__" not in line
            ):
                migrations.append(Migration(os.path.join(self.django_path, line)))
        diff_process.wait()

        if diff_process.returncode != 0:
            output = []
            for line in map(clean_bytes_to_str, diff_process.stderr.readlines()):
                output.append(line)
            logger.error("Error while git diff command:\n{}".format("".join(output)))
            raise Exception("Error while executing git diff command")
        return migrations

    def _gather_all_migrations(self):
        migrations = []
        for root, dirs, files in os.walk(self.django_path):
            for file_name in sorted(files):
                if (
                    os.path.basename(root) == MIGRATION_FOLDER_NAME
                    and file_name.endswith(".py")
                    and file_name != "__init__.py"
                ):
                    full_migration_path = os.path.join(root, file_name)
                    migrations.append(Migration(full_migration_path))
        return migrations

    def should_ignore_migration(self, app_name, migration_name):
        return (
            (self.include_apps and app_name not in self.include_apps)
            or (self.exclude_apps and app_name in self.exclude_apps)
            or (
                self.ignore_name_contains
                and self.ignore_name_contains in migration_name
            )
            or (migration_name in self.ignore_name)
        )


def _main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Detect backward incompatible django migrations."
    )
    parser.add_argument(
        "django_folder",
        metavar="DJANGO_FOLDER",
        type=str,
        nargs=1,
        help="the path to the django project",
    )
    parser.add_argument(
        "commit_id",
        metavar="GIT_COMMIT_ID",
        type=str,
        nargs="?",
        help=(
            "if specified, only migrations since this commit "
            "will be taken into account. If not specified, "
            "the initial repo commit will be used"
        ),
    )
    parser.add_argument(
        "--ignore-name-contains",
        type=str,
        nargs="?",
        help="ignore migrations containing this name",
    )
    parser.add_argument(
        "--ignore-name",
        type=str,
        nargs="*",
        help="ignore migrations with exactly one of these names",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="print more information during execution",
    )
    parser.add_argument(
        "--database",
        type=str,
        nargs="?",
        help=(
            "specify the database for which to generate the SQL. " "Defaults to default"
        ),
    )

    cache_group = parser.add_mutually_exclusive_group(required=False)
    cache_group.add_argument(
        "--cache-path",
        type=str,
        help="specify a directory that should be used to" "store cache-files in.",
    )
    cache_group.add_argument(
        "--no-cache", action="store_true", help="don't use a cache"
    )

    incl_excl_group = parser.add_mutually_exclusive_group(required=False)
    incl_excl_group.add_argument(
        "--include-apps",
        type=str,
        nargs="*",
        help="check only migrations that are in the specified django apps",
    )
    incl_excl_group.add_argument(
        "--exclude-apps",
        type=str,
        nargs="*",
        help="ignore migrations that are in the specified django apps",
    )

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(format="%(message)s", level=logging.DEBUG)
    else:
        logging.basicConfig(format="%(message)s")

    folder_name = args.django_folder[0]
    # Create and use linter
    linter = MigrationLinter(
        folder_name,
        ignore_name_contains=args.ignore_name_contains,
        ignore_name=args.ignore_name,
        include_apps=args.include_apps,
        exclude_apps=args.exclude_apps,
        database=args.database,
        cache_path=args.cache_path,
        no_cache=args.no_cache,
    )
    linter.lint_all_migrations(git_commit_id=args.commit_id)
    linter.print_summary()
    if linter.has_errors:
        sys.exit(1)


if __name__ == "__main__":
    _main()



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from django_migration_linter import MigrationLinter
from tests import fixtures


class BackwardcompatibilityDetectionTest(unittest.TestCase):
    def tearDown(self, *args, **kwargs):
        fixtures.clear_all_git_projects()
        super(
            BackwardcompatibilityDetectionTest,
            self).tearDown(*args, **kwargs)

    def _test_linter_finds_errors(self, path, commit_id=None):
        linter = MigrationLinter(path)
        linter.lint_all_migrations(git_commit_id=commit_id)
        self.assertTrue(linter.has_errors)

    def _test_linter_finds_no_errors(self, path, commit_id=None):
        linter = MigrationLinter(path)
        linter.lint_all_migrations(git_commit_id=commit_id)
        self.assertFalse(linter.has_errors)

    # *** Tests ***
    def test_create_table_with_not_null_column(self):
        test_project_path = fixtures.CREATE_TABLE_WITH_NOT_NULL_COLUMN_PROJECT
        self._test_linter_finds_no_errors(test_project_path)

    def test_detect_adding_not_null_column(self):
        test_project_path = fixtures.ADD_NOT_NULL_COLUMN_PROJECT
        self._test_linter_finds_errors(test_project_path)

    def test_detect_drop_column(self):
        test_project_path = fixtures.DROP_COLUMN_PROJECT
        self._test_linter_finds_errors(test_project_path)

    def test_detect_rename_column(self):
        test_project_path = fixtures.RENAME_COLUMN_PROJECT
        self._test_linter_finds_errors(test_project_path)

    def test_detect_rename_table(self):
        test_project_path = fixtures.RENAME_TABLE_PROJECT
        self._test_linter_finds_errors(test_project_path)

    def test_ignore_migration(self):
        test_project_path = fixtures.IGNORE_MIGRATION_PROJECT
        self._test_linter_finds_no_errors(test_project_path)

    def test_accept_not_null_column_followed_by_adding_default(self):
        test_project_path = \
            fixtures.ADD_NOT_NULL_COLUMN_FOLLOWED_BY_DEFAULT_PROJECT
        self._test_linter_finds_no_errors(test_project_path)

    def test_detect_alter_column(self):
        test_project_path = fixtures.ALTER_COLUMN_PROJECT
        self._test_linter_finds_errors(test_project_path)

    def test_no_specify_git_hash(self):
        test_project_path = fixtures.MULTI_COMMIT_PROJECT
        fixtures.prepare_git_project(test_project_path)
        self._test_linter_finds_errors(test_project_path, commit_id=None)

    def test_specify_git_hash_by_commit_hash(self):
        test_project_path = fixtures.MULTI_COMMIT_PROJECT
        fixtures.prepare_git_project(test_project_path)
        self._test_linter_finds_no_errors(
            test_project_path,
            commit_id='d7125d5f4f0cc9623f670a66c54f131acc50032d')

    #def test_specify_git_hash_by_tag(self):
    #    test_project_path = fixtures.MULTI_COMMIT_PROJECT
    #    fixtures.prepare_git_project(test_project_path)
    #    self._test_linter_finds_no_errors(test_project_path, commit_id='tag1')



from django_migration_linter.utils import split_migration_path


class Migration(object):
    def __init__(self, abs_path):
        self.abs_path = abs_path
        self.app_name, self.name = split_migration_path(self.abs_path)

    def __lt__(self, other):
        if not isinstance(other, Migration):
            return NotImplemented
        return self.abs_path < other.abs_path



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pprint
import shutil
import tempfile
import unittest
import sys

from tests import fixtures

from django_migration_linter import MigrationLinter, Cache, DEFAULT_CACHE_PATH

if sys.version_info >= (3, 3):
    import unittest.mock as mock
else:
    import mock


class CacheTest(unittest.TestCase):
    MIGRATION_FILE = os.path.join(fixtures.ALTER_COLUMN_PROJECT, 'test_app', 'migrations', '0001_initial.py')
    fd, temp_path = tempfile.mkstemp()

    def setUp(self):
        shutil.copy2(self.MIGRATION_FILE, self.temp_path)

    def tearDown(self):
        shutil.copy2(self.temp_path, self.MIGRATION_FILE)
        os.remove(self.temp_path)

    def test_cache_normal(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_project_add_not_null_column.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)
        linter = MigrationLinter(fixtures.ADD_NOT_NULL_COLUMN_PROJECT)

        with mock.patch.object(MigrationLinter, 'get_sql', wraps=linter.get_sql)as sql_mock:
            linter.lint_all_migrations()
            self.assertEqual(sql_mock.call_count, 2)

        cache = Cache(
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT,
            DEFAULT_CACHE_PATH
        )
        cache.load()

        self.assertEqual(cache['3ef74e7f3e53e273e2fc95379248d58d']['result'], 'OK')
        self.assertEqual(cache['e1e312b6d08ecbe017c25c58fc2be257']['result'], 'ERR')
        self.assertListEqual(
            cache['e1e312b6d08ecbe017c25c58fc2be257']['errors'],
            [{'err_msg': 'RENAMING tables', 'code': 'RENAME_TABLE', 'table': None, 'column': None}]
        )

        # Start the Linter again -> should use cache now.
        linter = MigrationLinter(fixtures.ADD_NOT_NULL_COLUMN_PROJECT)

        with mock.patch.object(MigrationLinter, 'get_sql', wraps=linter.get_sql) as sql_mock:
            linter.lint_all_migrations()
            self.assertEqual(sql_mock.call_count, 0)

        self.assertTrue(linter.has_errors)

    def test_cache_ignored(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_project_ignore_migration.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)
        linter = MigrationLinter(fixtures.IGNORE_MIGRATION_PROJECT)

        with mock.patch.object(MigrationLinter, 'get_sql', wraps=linter.get_sql) as sql_mock:
            linter.lint_all_migrations()
            self.assertEqual(sql_mock.call_count, 2)

        cache = Cache(
            fixtures.IGNORE_MIGRATION_PROJECT,
            DEFAULT_CACHE_PATH
        )
        cache.load()

        self.assertEqual(cache['63230606af0eccaef7f1f78c537c624c']['result'], 'OK')
        self.assertEqual(cache['5c5ca1780a9f28439c1defc1f32af894']['result'], 'IGNORE')

        # Start the Linter again -> should use cache now.
        linter = MigrationLinter(fixtures.IGNORE_MIGRATION_PROJECT)

        with mock.patch.object(MigrationLinter, 'get_sql', wraps=linter.get_sql) as sql_mock:
            linter.lint_all_migrations()
            self.assertEqual(sql_mock.call_count, 0)

    def test_cache_ignored_command_line(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_project_ignore_migration.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)
        linter = MigrationLinter(fixtures.IGNORE_MIGRATION_PROJECT,
                                 ignore_name_contains='0001')

        with mock.patch.object(MigrationLinter, 'get_sql', wraps=linter.get_sql) as sql_mock:
            linter.lint_all_migrations()
            self.assertEqual(sql_mock.call_count, 1)

        cache = Cache(
            fixtures.IGNORE_MIGRATION_PROJECT,
            DEFAULT_CACHE_PATH
        )
        cache.load()

        self.assertNotIn('63230606af0eccaef7f1f78c537c624c', cache)
        self.assertEqual(cache['5c5ca1780a9f28439c1defc1f32af894']['result'], 'IGNORE')

    def test_cache_modified(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_project_alter_column.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)
        linter = MigrationLinter(fixtures.ALTER_COLUMN_PROJECT)
        linter.lint_all_migrations()
        cache = Cache(
            fixtures.ALTER_COLUMN_PROJECT,
            DEFAULT_CACHE_PATH
        )
        cache.load()

        self.assertEqual(cache['8589aa107b6da296c4b49cd2681d2230']['result'], 'OK',
                         'If this fails, tearDown might have failed to remove '
                         'the modification from tests/test_project_fixtures/'
                         'test_project_alter_column/test_app/migrations/0001_initial.py'
                         )
        self.assertEqual(cache['8f54c4a434cfaa9838e8ca12eb988255']['result'], 'ERR')
        self.assertListEqual(
            cache['8f54c4a434cfaa9838e8ca12eb988255']['errors'],
            [{u'err_msg': u'ALTERING columns (Could be backward compatible. '
                          u'You may ignore this migration.)',
              u'code': u'ALTER_COLUMN',
              u'table': u'test_app_a',
              u'column': None}
             ]
        )

        # Modify migration
        with open(self.MIGRATION_FILE, "a") as f:
            f.write("# modification at the end of the file\n")

        # Start the Linter again -> Cache should look different now
        linter = MigrationLinter(fixtures.ALTER_COLUMN_PROJECT)
        linter.lint_all_migrations()
        cache = Cache(
            fixtures.ALTER_COLUMN_PROJECT,
            DEFAULT_CACHE_PATH
        )
        cache.load()

        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(cache)

        self.assertNotIn('8589aa107b6da296c4b49cd2681d2230', cache)
        self.assertEqual(cache['44f9d6d019d6f158946c2819b2bd8bf9']['result'], 'OK')



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest

from tests import fixtures
from django_migration_linter.utils import is_django_project, is_directory, find_project_settings_module, \
    split_path, split_migration_path


class UtilityFunctionTest(unittest.TestCase):
    def test_detect_django_project(self):
        self.assertTrue(is_django_project(fixtures.ADD_NOT_NULL_COLUMN_PROJECT))

    def test_detect_not_django_project(self):
        self.assertFalse(is_django_project(fixtures._BASE_DIR))

    def test_detect_directory(self):
        self.assertTrue(is_directory(fixtures.ADD_NOT_NULL_COLUMN_PROJECT))

    def test_detect_not_directory(self):
        self.assertFalse(is_directory(
            os.path.join(fixtures.ADD_NOT_NULL_COLUMN_PROJECT, 'manage.py')))

    def test_find_project_settings_module_same_project_name(self):
        expected = 'test_project_add_not_null_column.settings'
        actual = find_project_settings_module(fixtures.ADD_NOT_NULL_COLUMN_PROJECT)
        self.assertEqual(actual, expected)

    def test_find_project_settings_module_different_project_name(self):
        expected = 'test_project.settings'
        actual = find_project_settings_module(fixtures.CORRECT_PROJECT)
        self.assertEqual(actual, expected)

    def test_split_path(self):
        splitted = split_path('foo/bar/fuz.py')
        self.assertEqual(len(splitted), 3)
        self.assertEqual(splitted[0], 'foo')
        self.assertEqual(splitted[1], 'bar')
        self.assertEqual(splitted[2], 'fuz.py')

    def test_split_full_path(self):
        splitted = split_path('/foo/bar/fuz.py')
        self.assertEqual(len(splitted), 3)
        self.assertEqual(splitted[0], 'foo')
        self.assertEqual(splitted[1], 'bar')
        self.assertEqual(splitted[2], 'fuz.py')

    def test_split_migration_long_path(self):
        input_path = 'apps/the_app/migrations/0001_stuff.py'
        app, mig = split_migration_path(input_path)
        self.assertEqual(app, 'the_app')
        self.assertEqual(mig, '0001_stuff')

    def test_split_migration_path(self):
        input_path = 'the_app/migrations/0001_stuff.py'
        app, mig = split_migration_path(input_path)
        self.assertEqual(app, 'the_app')
        self.assertEqual(mig, '0001_stuff')

    def test_split_migration_full_path(self):
        input_path = '/home/user/djangostuff/apps/the_app/migrations/0001_stuff.py'
        app, mig = split_migration_path(input_path)
        self.assertEqual(app, 'the_app')
        self.assertEqual(mig, '0001_stuff')



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import sys

from .constants import MIGRATION_FOLDER_NAME


def is_django_project(path):
    django_manage_file = os.path.join(path, "manage.py")
    return os.path.isfile(django_manage_file)


def is_directory(path):
    return os.path.isdir(path)


def print_stderr(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def find_project_settings_module(path):
    """Explore path recursively to the first settings.py file
    and translate the name to package notation 'mysite.settings'
    """
    for root, dirs, files in os.walk(path):
        for file_name in files:
            if file_name == "settings.py":
                return (
                    os.path.join(root.replace(path, ""), file_name)
                    .replace("/", ".")
                    .rstrip(".py")
                )


def split_path(path):
    a, b = os.path.split(path)
    return (split_path(a) if (len(a) > 0 and a != "/") else []) + [b]


def split_migration_path(migration_path):
    decomposed_path = split_path(migration_path)
    for i, p in enumerate(decomposed_path):
        if p == MIGRATION_FOLDER_NAME:
            return (decomposed_path[i - 1], os.path.splitext(decomposed_path[i + 1])[0])


def compose_migration_path(django_folder, app_name, migration):
    return os.path.join(
        django_folder, app_name, MIGRATION_FOLDER_NAME, "{0}.py".format(migration)
    )


def clean_bytes_to_str(byte_input):
    return byte_input.decode("utf-8").strip()



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from subprocess import Popen, PIPE
from django_migration_linter import utils, DEFAULT_CACHE_PATH
from tests import fixtures
import sys


class CallLinterFromCommandLineTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(CallLinterFromCommandLineTest, cls).setUpClass()
        linter_name = 'django-migration-linter'
        cls.linter_exec = '{0}/bin/{1}'.format(sys.prefix,  linter_name) if hasattr(sys, 'real_prefix') else linter_name

    def test_call_linter_cmd_line_working(self):
        cmd = '{0} --no-cache {1}'.format(self.linter_exec, fixtures.CORRECT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 6)
        self.assertEqual(
            sorted(lines[:4]),
            sorted([
                 "(test_app1, 0001_initial)... OK",
                 "(test_app1, 0002_a_new_null_field)... OK",
                 "(test_app2, 0001_foo)... OK",
                 "(test_app3, 0001_initial)... OK",
             ])
        )

    def test_call_linter_cmd_line_errors(self):
        cmd = '{0} --no-cache {1}'.format(
            self.linter_exec,
            fixtures.ADD_NOT_NULL_COLUMN_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertNotEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 5)
        self.assertTrue(lines[0].endswith('OK'))
        self.assertTrue(lines[1].endswith('ERR'))
        self.assertTrue('RENAMING tables' in lines[2])

    def test_call_linter_cmd_line_exclude_apps(self):
        cmd = '{0} --no-cache {1} --exclude-apps test_app2'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 6)
        self.assertEqual(
            sorted(lines[:4]),
            sorted([
                 "(test_app1, 0001_initial)... OK",
                 "(test_app1, 0002_a_new_null_field)... OK",
                 "(test_app2, 0001_foo)... IGNORE",
                 "(test_app3, 0001_initial)... OK",
             ])
        )

    def test_call_linter_cmd_line_include_apps(self):
        cmd = '{0} --no-cache {1} --include-apps test_app2'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 6)
        self.assertEqual(
            sorted(lines[:4]),
            sorted([
                 "(test_app1, 0001_initial)... IGNORE",
                 "(test_app1, 0002_a_new_null_field)... IGNORE",
                 "(test_app2, 0001_foo)... OK",
                 "(test_app3, 0001_initial)... IGNORE",
             ])
        )

    def test_call_linter_cmd_line_ignore_name(self):
        cmd = '{0} --no-cache {1} --ignore-name 0001_initial'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 6)
        self.assertEqual(
            sorted(lines[:4]),
            sorted([
                 "(test_app1, 0001_initial)... IGNORE",
                 "(test_app1, 0002_a_new_null_field)... OK",
                 "(test_app2, 0001_foo)... OK",
                 "(test_app3, 0001_initial)... IGNORE",
             ])
        )

    def test_call_linter_cmd_line_ignore_name_contains(self):
        cmd = '{0} --no-cache {1} --ignore-name-contains 0001'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 6)
        self.assertEqual(
            sorted(lines[:4]),
            sorted([
                 "(test_app1, 0001_initial)... IGNORE",
                 "(test_app1, 0002_a_new_null_field)... OK",
                 "(test_app2, 0001_foo)... IGNORE",
                 "(test_app3, 0001_initial)... IGNORE",
             ])
        )

    def test_call_linter_cmd_line_cache(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_correct_project.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)
        cmd = '{0} {1}'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT
        )

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertTrue(os.path.exists(DEFAULT_CACHE_PATH))

    def test_call_linter_cmd_line_cache_path(self):
        if os.path.exists('/tmp/migration-linter-cache-tests/'):
            shutil.rmtree('/tmp/migration-linter-cache-tests/')
        cmd = '{0} {1} --cache-path={2}'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT,
            '/tmp/migration-linter-cache-tests/'
        )

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        self.assertTrue(os.path.exists('/tmp/migration-linter-cache-tests'))

    def test_call_linter_cmd_line_no_cache(self):
        cache_file = os.path.join(DEFAULT_CACHE_PATH, 'test_correct_project.pickle')
        if os.path.exists(cache_file):
            os.remove(cache_file)

        cmd = '{0} --no-cache {1}'.format(
            self.linter_exec,
            fixtures.CORRECT_PROJECT
        )

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        self.assertFalse(os.path.exists(cache_file))

    def test_call_linter_cmd_line_git_id(self):
        cmd = '{0} --no-cache {1} d7125d5f4f0cc9623f670a66c54f131acc50032d'.format(
            self.linter_exec,
            fixtures.MULTI_COMMIT_PROJECT)
        fixtures.prepare_git_project(fixtures.MULTI_COMMIT_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 3)
        self.assertTrue(lines[0].endswith('OK'))
        self.assertTrue(lines[1].startswith('*** Summary'))

    def test_call_linter_with_deleted_migrations(self):
        cmd = '{0} --no-cache {1} 154ecf6119325cc3b1f3f5a4e709bfbd61a4a4ba'.format(
            self.linter_exec,
            fixtures.DELETED_MIGRATION_PROJECT)
        fixtures.prepare_git_project(fixtures.DELETED_MIGRATION_PROJECT)

        process = Popen(
            cmd, shell=True, stdout=PIPE, stderr=PIPE)
        process.wait()
        self.assertEqual(process.returncode, 0)
        lines = list(map(utils.clean_bytes_to_str, process.stdout.readlines()))
        self.assertEqual(len(lines), 3)
        self.assertTrue(lines[0].endswith('OK'))
        self.assertTrue(lines[1].startswith('*** Summary'))



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

from django_migration_linter import MigrationLinter
from django_migration_linter.migration import Migration
from tests import fixtures


class TestMultipleLinters(unittest.TestCase):
    def test_multiple_linters(self):
        l1 = MigrationLinter(fixtures.ADD_NOT_NULL_COLUMN_PROJECT)
        l2 = MigrationLinter(fixtures.RENAME_COLUMN_PROJECT)
        l3 = MigrationLinter(fixtures.CORRECT_PROJECT)

        l1.lint_migration(Migration(os.path.join(fixtures.ADD_NOT_NULL_COLUMN_PROJECT, 'test_app/migrations/0002_add_new_not_null_field.py')))
        l2.lint_migration(Migration(os.path.join(fixtures.RENAME_COLUMN_PROJECT, 'test_app/migrations/0002_rename_column.py')))
        l3.lint_migration(Migration(os.path.join(fixtures.CORRECT_PROJECT, 'test_app1/migrations/0001_initial.py')))
        l3.lint_migration(Migration(os.path.join(fixtures.CORRECT_PROJECT, 'test_app1/migrations/0002_a_new_null_field.py')))

        self.assertTrue(l1.has_errors)
        self.assertTrue(l2.has_errors)
        self.assertFalse(l3.has_errors)

        self.assertEqual(l1.nb_total, 1)
        self.assertEqual(l2.nb_total, 1)
        self.assertEqual(l3.nb_total, 2)



# Copyright 2019 3YOURMIND GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import re
from os import path
from setuptools import setup, find_packages


PROJECT_DIR = path.abspath(path.dirname(__file__))


def get_version():
    constants = path.join(PROJECT_DIR, 'django_migration_linter', 'constants.py')
    _version_re = re.compile(r"__version__\s+=\s+(?P<version>.*)")
    with open(constants, "r") as f:
        match = _version_re.search(f.read())
        version = match.group("version") if match is not None else '"unknown"'
    return str(ast.literal_eval(version))


with open(path.join(PROJECT_DIR, 'README.rst')) as f:
    long_description = f.read()

install_requirements = [
    'django>=1.11',
    'appdirs==1.4.3'
]

test_requirements = [
    'pytest',
    'tox',
    'django-fake-database-backends',
]

# noinspection PyUnresolvedReferences
setup(
    name='django-migration-linter',
    version=get_version(),

    description='Detect backward incompatible migrations for your django project',
    long_description=long_description,
    url='https://github.com/3YOURMIND/django-migration-linter',
    author='3YOURMIND GmbH',
    license='Apache License 2.0',

    packages=find_packages(exclude=['tests/']),
    install_requires=install_requirements,
    extras_require={
        'test': test_requirements
    },
    entry_points={
        'console_scripts': [
            'django-migration-linter = django_migration_linter.migration_linter:_main'
        ]
    },

    keywords='django migration lint linter database backward compatibility',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)
